@@ -6,7 +6,7 @@
  * (the "License"); you may not use this file except in compliance with
  * the License.  You may obtain a copy of the License at
  *
- *      http://www.apache.org/licenses/LICENSE-2.0
+ *      https://www.apache.org/licenses/LICENSE-2.0
  *
  * Unless required by applicable law or agreed to in writing, software
  * distributed under the License is distributed on an "AS IS" BASIS,
@@ -18,16 +18,22 @@
 package org.apache.commons.jexl3;
 
 import static org.junit.jupiter.api.Assertions.assertEquals;
+import static org.junit.jupiter.api.Assertions.assertFalse;
+import static org.junit.jupiter.api.Assertions.assertInstanceOf;
 import static org.junit.jupiter.api.Assertions.assertNotNull;
+import static org.junit.jupiter.api.Assertions.assertThrows;
 import static org.junit.jupiter.api.Assertions.assertTrue;
 import static org.junit.jupiter.api.Assertions.fail;
 
 import java.io.StringWriter;
+import java.math.MathContext;
 import java.text.DecimalFormat;
 import java.text.SimpleDateFormat;
+import java.time.Instant;
 import java.util.ArrayList;
 import java.util.Arrays;
 import java.util.Calendar;
+import java.util.Comparator;
 import java.util.Date;
 import java.util.HashMap;
 import java.util.HashSet;
@@ -39,6 +45,7 @@ import java.util.Set;
 import java.util.SortedSet;
 import java.util.TimeZone;
 import java.util.TreeSet;
+import java.util.regex.Pattern;
 
 import org.apache.commons.jexl3.junit.Asserter;
 import org.junit.jupiter.api.BeforeEach;
@@ -48,7 +55,7 @@ import org.junit.jupiter.api.Test;
  * Tests for the startsWith, endsWith, match and range operators.
  */
 @SuppressWarnings({"UnnecessaryBoxing", "AssertEqualsBetweenInconvertibleTypes"})
-public class ArithmeticOperatorTest extends JexlTestCase {
+class ArithmeticOperatorTest extends JexlTestCase {
     public static class Aggregate {
         public static int sum(final Iterable<Integer> ii) {
             int sum = 0;
@@ -76,14 +83,15 @@ public class ArithmeticOperatorTest extends JexlTestCase {
             try {
                 final Calendar cal = Calendar.getInstance(UTC);
                 cal.setTime(date);
-                if ("yyyy".equals(key)) {
+                switch (key) {
+                case "yyyy":
                     return cal.get(Calendar.YEAR);
-                }
-                if ("MM".equals(key)) {
+                case "MM":
                     return cal.get(Calendar.MONTH) + 1;
-                }
-                if ("dd".equals(key)) {
+                case "dd":
                     return cal.get(Calendar.DAY_OF_MONTH);
+                default:
+                    break;
                 }
                 // Otherwise treat as format mask
                 final SimpleDateFormat df = new SimpleDateFormat(key); //, dfs);
@@ -110,15 +118,21 @@ public class ArithmeticOperatorTest extends JexlTestCase {
             return setDateValue(date, identifier, value);
         }
 
-        protected Object setDateValue(final Date date, final String key, final Object value) throws Exception {
+        protected Object setDateValue(final Date date, final String key, final Object value) {
             final Calendar cal = Calendar.getInstance(UTC);
             cal.setTime(date);
-            if ("yyyy".equals(key)) {
+            switch (key) {
+            case "yyyy":
                 cal.set(Calendar.YEAR, toInteger(value));
-            } else if ("MM".equals(key)) {
+                break;
+            case "MM":
                 cal.set(Calendar.MONTH, toInteger(value) - 1);
-            } else if ("dd".equals(key)) {
+                break;
+            case "dd":
                 cal.set(Calendar.DAY_OF_MONTH, toInteger(value));
+                break;
+            default:
+                break;
             }
             date.setTime(cal.getTimeInMillis());
             return date;
@@ -234,24 +248,24 @@ public class ArithmeticOperatorTest extends JexlTestCase {
         asserter.setStrict(false);
     }
 
-    @Test public void test373a() {
+    @Test void test373a() {
         testSelfAssignOperators("y.add(x++)", 42, 42, 43);
     }
 
-    @Test public void test373b() {
+    @Test void test373b() {
         testSelfAssignOperators("y.add(++x)", 42, 43, 43);
     }
 
-    @Test public void test373c() {
+    @Test void test373c() {
         testSelfAssignOperators("y.add(x--)", 42, 42, 41);
     }
 
-    @Test public void test373d() {
+    @Test void test373d() {
         testSelfAssignOperators("y.add(--x)", 42, 41, 41);
     }
 
     @Test
-    public void test391() throws Exception {
+    void test391() throws Exception {
         // with literals
         for(final String src : Arrays.asList(
                 "2 =~ [1, 2, 3, 4]",
@@ -278,7 +292,7 @@ public class ArithmeticOperatorTest extends JexlTestCase {
     }
 
     @Test
-    public void testDateArithmetic() throws Exception {
+    void testDateArithmetic() {
         final Date d = new Date();
         final JexlContext jc = new MapContext();
         final JexlEngine jexl = new JexlBuilder().cache(32).arithmetic(new DateArithmetic(true)).create();
@@ -293,7 +307,7 @@ public class ArithmeticOperatorTest extends JexlTestCase {
     }
 
     @Test
-    public void testFormatArithmetic() throws Exception {
+    void testFormatArithmetic() {
         final Calendar cal = Calendar.getInstance(UTC);
         cal.set(1969, Calendar.AUGUST, 20);
         final Date x0 = cal.getTime();
@@ -337,7 +351,7 @@ public class ArithmeticOperatorTest extends JexlTestCase {
     }
 
     @Test
-    public void testFormatArithmeticJxlt() throws Exception {
+    void testFormatArithmeticJxlt() throws Exception {
         final Map<String, Object> ns = new HashMap<>();
         ns.put("calc", Aggregate.class);
         final Calendar cal = Calendar.getInstance(UTC);
@@ -374,7 +388,7 @@ public class ArithmeticOperatorTest extends JexlTestCase {
     }
 
     @Test
-    public void testIncrementOperatorOnNull() throws Exception {
+    void testIncrementOperatorOnNull() {
         final JexlEngine jexl = new JexlBuilder().strict(false).create();
         JexlScript script;
         Object result;
@@ -388,7 +402,7 @@ public class ArithmeticOperatorTest extends JexlTestCase {
 
     @Test
     @SuppressWarnings("unchecked")
-    public void testInterval() throws Exception {
+    void testInterval() {
         final Map<String, Object> ns = new HashMap<>();
         ns.put("calc", Aggregate.class);
         final JexlEngine jexl = new JexlBuilder().namespaces(ns).create();
@@ -397,7 +411,7 @@ public class ArithmeticOperatorTest extends JexlTestCase {
 
         script = jexl.createScript("1 .. 3");
         result = script.execute(null);
-        assertTrue(result instanceof Iterable<?>);
+        assertInstanceOf(Iterable.class, result);
         Iterator<Integer> ii = ((Iterable<Integer>) result).iterator();
         assertEquals(Integer.valueOf(1), ii.next());
         assertEquals(Integer.valueOf(2), ii.next());
@@ -405,7 +419,7 @@ public class ArithmeticOperatorTest extends JexlTestCase {
 
         script = jexl.createScript("(4 - 3) .. (9 / 3)");
         result = script.execute(null);
-        assertTrue(result instanceof Iterable<?>);
+        assertInstanceOf(Iterable.class, result);
         ii = ((Iterable<Integer>) result).iterator();
         assertEquals(Integer.valueOf(1), ii.next());
         assertEquals(Integer.valueOf(2), ii.next());
@@ -426,7 +440,7 @@ public class ArithmeticOperatorTest extends JexlTestCase {
     }
 
     @Test
-    public void testMatch() throws Exception {
+    void testMatch() throws Exception {
         // check in/not-in on array, list, map, set and duck-type collection
         final int[] ai = {2, 4, 42, 54};
         final List<Integer> al = new ArrayList<>();
@@ -443,8 +457,8 @@ public class ArithmeticOperatorTest extends JexlTestCase {
         final Set<Integer> as = ad.values;
         final Object[] vars = {ai, al, am, ad, as, ic};
 
-        for (final Object var : vars) {
-            asserter.setVariable("container", var);
+        for (final Object variable : vars) {
+            asserter.setVariable("container", variable);
             for (final int x : ai) {
                 asserter.setVariable("x", x);
                 asserter.assertExpression("x =~ container", Boolean.TRUE);
@@ -455,7 +469,7 @@ public class ArithmeticOperatorTest extends JexlTestCase {
     }
 
     @Test
-    public void testNotStartsEndsWith() throws Exception {
+    void testNotStartsEndsWith() throws Exception {
         asserter.setVariable("x", "foobar");
         asserter.assertExpression("x !^ 'foo'", Boolean.FALSE);
         asserter.assertExpression("x !$ 'foo'", Boolean.TRUE);
@@ -475,7 +489,7 @@ public class ArithmeticOperatorTest extends JexlTestCase {
     }
 
     @Test
-    public void testNotStartsEndsWithString() throws Exception {
+    void testNotStartsEndsWithString() throws Exception {
         asserter.setVariable("x", "foobar");
         asserter.assertExpression("x !^ 'foo'", Boolean.FALSE);
         asserter.assertExpression("x !$ 'foo'", Boolean.TRUE);
@@ -485,7 +499,7 @@ public class ArithmeticOperatorTest extends JexlTestCase {
     }
 
     @Test
-    public void testNotStartsEndsWithStringBuilder() throws Exception {
+    void testNotStartsEndsWithStringBuilder() throws Exception {
         asserter.setVariable("x", new StringBuilder("foobar"));
         asserter.assertExpression("x !^ 'foo'", Boolean.FALSE);
         asserter.assertExpression("x !$ 'foo'", Boolean.TRUE);
@@ -495,7 +509,7 @@ public class ArithmeticOperatorTest extends JexlTestCase {
     }
 
     @Test
-    public void testNotStartsEndsWithStringDot() throws Exception {
+    void testNotStartsEndsWithStringDot() throws Exception {
         asserter.setVariable("x.y", "foobar");
         asserter.assertExpression("x.y !^ 'foo'", Boolean.FALSE);
         asserter.assertExpression("x.y !$ 'foo'", Boolean.TRUE);
@@ -505,12 +519,12 @@ public class ArithmeticOperatorTest extends JexlTestCase {
     }
 
     @Test
-    public void testOperatorError() throws Exception {
-        testOperatorError(true);
-        testOperatorError(false);
+    void testOperatorError() throws Exception {
+        runOperatorError(true);
+        runOperatorError(false);
     }
 
-    private void testOperatorError(final boolean silent) throws Exception {
+    private void runOperatorError(final boolean silent) {
         final CaptureLog log = new CaptureLog();
         final DateContext jc = new DateContext();
         final Date d = new Date();
@@ -533,7 +547,7 @@ public class ArithmeticOperatorTest extends JexlTestCase {
     }
 
     @Test
-    public void testRegexp() throws Exception {
+    void testRegexp() throws Exception {
         asserter.setVariable("str", "abc456");
         asserter.assertExpression("str =~ '.*456'", Boolean.TRUE);
         asserter.assertExpression("str !~ 'ABC.*'", Boolean.TRUE);
@@ -563,7 +577,7 @@ public class ArithmeticOperatorTest extends JexlTestCase {
     }
 
     @Test
-    public void testRegexp2() throws Exception {
+    void testRegexp2() throws Exception {
         asserter.setVariable("str", "abc456");
         asserter.assertExpression("str =~ ~/.*456/", Boolean.TRUE);
         asserter.assertExpression("str !~ ~/ABC.*/", Boolean.TRUE);
@@ -586,7 +600,7 @@ public class ArithmeticOperatorTest extends JexlTestCase {
         assertEquals(y0, y.get(0), "y0");
     }
     @Test
-    public void testStartsEndsWith() throws Exception {
+    void testStartsEndsWith() throws Exception {
         asserter.setVariable("x", "foobar");
         asserter.assertExpression("x =^ 'foo'", Boolean.TRUE);
         asserter.assertExpression("x =$ 'foo'", Boolean.FALSE);
@@ -605,7 +619,7 @@ public class ArithmeticOperatorTest extends JexlTestCase {
         asserter.assertExpression("x =^ [42, 54]", Boolean.TRUE);
     }
     @Test
-    public void testStartsEndsWithString() throws Exception {
+    void testStartsEndsWithString() throws Exception {
         asserter.setVariable("x", "foobar");
         asserter.assertExpression("x =^ 'foo'", Boolean.TRUE);
         asserter.assertExpression("x =$ 'foo'", Boolean.FALSE);
@@ -615,7 +629,7 @@ public class ArithmeticOperatorTest extends JexlTestCase {
     }
 
     @Test
-    public void testStartsEndsWithStringBuilder() throws Exception {
+    void testStartsEndsWithStringBuilder() throws Exception {
         asserter.setVariable("x", new StringBuilder("foobar"));
         asserter.assertExpression("x =^ 'foo'", Boolean.TRUE);
         asserter.assertExpression("x =$ 'foo'", Boolean.FALSE);
@@ -625,7 +639,7 @@ public class ArithmeticOperatorTest extends JexlTestCase {
     }
 
     @Test
-    public void testStartsEndsWithStringDot() throws Exception {
+    void testStartsEndsWithStringDot() throws Exception {
         asserter.setVariable("x.y", "foobar");
         asserter.assertExpression("x.y =^ 'foo'", Boolean.TRUE);
         asserter.assertExpression("x.y =$ 'foo'", Boolean.FALSE);
@@ -634,4 +648,223 @@ public class ArithmeticOperatorTest extends JexlTestCase {
         asserter.assertExpression("x.y =$ 'foo'", Boolean.TRUE);
     }
 
+    /**
+     * A comparator using an evaluated expression on objects as comparison arguments.
+     * <p>Lifetime is the sort method; it is thus safe to encapsulate the context</p>
+     */
+     private static class PropertyComparator implements JexlCache.Reference, Comparator<Object> {
+        private final JexlContext context = JexlEngine.getThreadContext();
+        private final JexlArithmetic arithmetic;
+        private final JexlOperator.Uberspect operator;
+        private final JexlScript expr;
+        private Object cache;
+
+        PropertyComparator(final JexlArithmetic jexla, final JexlScript expr) {
+            this.arithmetic = jexla;
+            this.operator = JexlEngine.getThreadEngine().getUberspect().getOperator(arithmetic);
+            this.expr = expr;
+        }
+        @Override
+        public int compare(final Object o1, final Object o2) {
+            final Object left = expr.execute(context, o1);
+            final Object right = expr.execute(context, o2);
+            final Object result = operator.tryOverload(this, JexlOperator.COMPARE, left, right);
+            if (result instanceof Integer) {
+                return (int) result;
+            }
+            return arithmetic.compare(left, right, JexlOperator.COMPARE);
+        }
+
+        @Override
+        public Object getCache() {
+            return cache;
+        }
+
+        @Override
+        public void setCache(final Object cache) {
+            this.cache = cache;
+        }
+    }
+
+    public static class SortingArithmetic extends JexlArithmetic {
+        public SortingArithmetic(final boolean strict) {
+            this(strict, null, Integer.MIN_VALUE);
+        }
+
+        private SortingArithmetic(final boolean strict, final MathContext context, final int scale) {
+            super(strict, context, scale);
+        }
+
+        public int compare(final Integer left, final Integer right) {
+            return left.compareTo(right);
+        }
+
+        public int compare(final String left, final String right) {
+            return left.compareTo(right);
+        }
+
+        /**
+         * Sorts an array using a script to evaluate the property used to compare elements.
+         *
+         * @param array the elements array
+         * @param expr  the property evaluation lambda
+         */
+        public void sort(final Object[] array, final JexlScript expr) {
+            Arrays.sort(array, new PropertyComparator(this, expr));
+        }
+    }
+
+    @Test
+    void testSortArray() {
+        final JexlEngine jexl = new JexlBuilder()
+                .cache(32)
+                .arithmetic(new SortingArithmetic(true))
+                .silent(false).create();
+        // test data, json like
+        final String src = "[{'id':1,'name':'John','type':9},{'id':2,'name':'Doe','type':7},{'id':3,'name':'Doe','type':10}]";
+        final Object a =  jexl.createExpression(src).evaluate(null);
+        assertNotNull(a);
+        // row 0 and 1 are not ordered
+        final Map[] m = (Map[]) a;
+        assertEquals(9, m[0].get("type"));
+        assertEquals(7, m[1].get("type"));
+        // sort the elements on the type
+        jexl.createScript("array.sort( e -> e.type )", "array").execute(null, a);
+        // row 0 and 1 are now ordered
+        assertEquals(7, m[0].get("type"));
+        assertEquals(9, m[1].get("type"));
+    }
+
+    public static class MatchingArithmetic extends JexlArithmetic {
+        public MatchingArithmetic(final boolean astrict) {
+            super(astrict);
+        }
+
+        public boolean contains(final Pattern[] container, final String str) {
+            for(final Pattern pattern : container) {
+                if (pattern.matcher(str).matches()) {
+                    return true;
+                }
+            }
+            return false;
+        }
+    }
+
+    @Test
+    void testPatterns() {
+        final JexlEngine jexl = new JexlBuilder().arithmetic(new MatchingArithmetic(true)).create();
+        final JexlScript script = jexl.createScript("str =~ [~/abc.*/, ~/def.*/]", "str");
+        assertTrue((boolean) script.execute(null, "abcdef"));
+        assertTrue((boolean) script.execute(null, "defghi"));
+        assertFalse((boolean) script.execute(null, "ghijkl"));
+    }
+
+    public static class Arithmetic428 extends JexlArithmetic {
+        public Arithmetic428(final boolean strict) {
+            this(strict, null, Integer.MIN_VALUE);
+        }
+
+        private Arithmetic428(final boolean strict, final MathContext context, final int scale) {
+            super(strict, context, scale);
+        }
+
+        public int compare(final Instant lhs, final String str) {
+            final Instant rhs = Instant.parse(str);
+            return lhs.compareTo(rhs);
+        }
+    }
+
+    static final List<Integer> LOOPS = new ArrayList<>(Arrays.asList(0, 1));
+
+    @Test
+    void test428() {
+        // see JEXL-428
+        final JexlEngine jexl = new JexlBuilder().cache(32).arithmetic(new Arithmetic428(true)).create();
+        final String rhsstr ="2024-09-09T10:42:42.00Z";
+        final Instant rhs = Instant.parse(rhsstr);
+        final String lhs = "2020-09-09T01:24:24.00Z";
+        JexlScript script;
+        script = jexl.createScript("x < y", "x", "y");
+        final JexlScript s0 = script;
+        assertThrows(JexlException.class, () -> s0.execute(null, 42, rhs));
+        for(final int i : LOOPS) { assertTrue((boolean) script.execute(null, lhs, rhs)); }
+        for(final int i : LOOPS) { assertTrue((boolean) script.execute(null, lhs, rhs)); }
+        for(final int i : LOOPS) { assertFalse((boolean) script.execute(null, rhs, lhs)); }
+        for(final int i : LOOPS) { assertFalse((boolean) script.execute(null, rhs, lhs)); }
+        for(final int i : LOOPS) { assertTrue((boolean) script.execute(null, lhs, rhs)); }
+        for(final int i : LOOPS) { assertFalse((boolean) script.execute(null, rhs, lhs)); }
+
+        script = jexl.createScript("x <= y", "x", "y");
+        final JexlScript s1 = script;
+        assertThrows(JexlException.class, () -> s1.execute(null, 42, rhs));
+        assertTrue((boolean) script.execute(null, lhs, rhs));
+        assertFalse((boolean) script.execute(null, rhs, lhs));
+
+        script = jexl.createScript("x >= y", "x", "y");
+        final JexlScript s2 = script;
+        assertThrows(JexlException.class, () -> s2.execute(null, 42, rhs));
+        assertFalse((boolean) script.execute(null, lhs, rhs));
+        assertFalse((boolean) script.execute(null, lhs, rhs));
+        assertTrue((boolean) script.execute(null, rhs, lhs));
+        assertTrue((boolean) script.execute(null, rhs, lhs));
+        assertFalse((boolean) script.execute(null, lhs, rhs));
+        assertTrue((boolean) script.execute(null, rhs, lhs));
+
+        script = jexl.createScript("x > y", "x", "y");
+        final JexlScript s3 = script;
+        assertThrows(JexlException.class, () -> s3.execute(null, 42, rhs));
+        assertFalse((boolean) script.execute(null, lhs, rhs));
+        assertTrue((boolean) script.execute(null, rhs, lhs));
+
+        script = jexl.createScript("x == y", "x", "y");
+        assertFalse((boolean) script.execute(null, 42, rhs));
+        assertFalse((boolean) script.execute(null, lhs, rhs));
+        assertFalse((boolean) script.execute(null, lhs, rhs));
+        assertTrue((boolean) script.execute(null, rhs, rhsstr));
+        assertTrue((boolean) script.execute(null, rhsstr, rhs));
+        assertFalse((boolean) script.execute(null, lhs, rhs));
+
+        script = jexl.createScript("x != y", "x", "y");
+        assertTrue((boolean) script.execute(null, 42, rhs));
+        assertTrue((boolean) script.execute(null, lhs, rhs));
+        assertFalse((boolean) script.execute(null, rhs, rhsstr));
+    }
+
+    public static class Arithmetic429 extends JexlArithmetic {
+        public Arithmetic429(final boolean astrict) {
+            super(astrict);
+        }
+
+        public int compare(final String lhs, final Number rhs) {
+            return lhs.compareTo(rhs.toString());
+        }
+    }
+
+    @Test
+    void test429a() {
+        final JexlEngine jexl = new JexlBuilder()
+                .arithmetic(new Arithmetic429(true))
+                .cache(32)
+                .create();
+        String src;
+        JexlScript script;
+        src = "'1.1' > 0";
+        script = jexl.createScript(src);
+        assertTrue((boolean) script.execute(null));
+        src = "1.2 <= '1.20'";
+        script = jexl.createScript(src);
+        assertTrue((boolean) script.execute(null));
+        src = "1.2 >= '1.2'";
+        script = jexl.createScript(src);
+        assertTrue((boolean) script.execute(null));
+        src = "1.2 < '1.2'";
+        script = jexl.createScript(src);
+        assertFalse((boolean) script.execute(null));
+        src = "1.2 > '1.2'";
+        script = jexl.createScript(src);
+        assertFalse((boolean) script.execute(null));
+        src = "1.20 == 'a'";
+        script = jexl.createScript(src);
+        assertFalse((boolean) script.execute(null));
+    }
 }
